/*
This source file is part of ProceduralGenerator (https://sourceforge.net/projects/proceduralgene/)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___Pg_BufferElement___
#define ___Pg_BufferElement___

namespace ProceduralTextures
{
	struct BufferElementDeclaration
	{
		unsigned int	m_uiStream;		//!< Stream index
		eELEMENT_USAGE	m_eUsage;		//!< Element usage
		eELEMENT_TYPE	m_eDataType;	//!< Element type
		size_t			m_uiOffset;		//!< Offset in stream
	};
	/**
	 *\brief		Function to initialise a \p BufferElementDeclaration
	 *\param[out]	p_declaration	The \p BufferElementDeclaration to initialise
	 *\param[in]	p_uiStream		The stream index of the buffer element
	 *\param[in]	p_eUsage		The usage of the buffer element
	 *\param[in]	p_eDataType		The buffer element data type
	 *\param[in]	p_uiOffset		The offset of the element in the buffer
	 */
	extern void Initialise( BufferElementDeclaration & p_declaration, unsigned int p_uiStream, eELEMENT_USAGE p_eUsage, eELEMENT_TYPE p_eDataType, size_t p_uiOffset = 0 );

	class BufferDeclaration
	{
	public:
		typedef std::vector< BufferElementDeclaration > BufferElementDeclarationArray;	//!< Typedef over an array of \p BufferElementDeclaration
		typedef BufferElementDeclarationArray::iterator 		iterator;				//!< Typedef over an iterator on \p BufferElementDeclarationArray
		typedef BufferElementDeclarationArray::const_iterator 	const_iterator;			//!< Typedef over a constant iterator on \p BufferElementDeclarationArray

	protected:
		BufferElementDeclarationArray m_arrayElements;	//!< Array of \p BufferElementDeclaration
		size_t m_uiStride;								//!< Size of 1 complete element

	public:
		/**
		 *\brief		Constructor, from a static array
		 *\param[in]	p_elements		The elements array
		 */
		template <size_t N> BufferDeclaration( const BufferElementDeclaration (& p_elements)[N] )
		{
			DoInitialise( p_elements, N);
		}
		/**
		 *\brief		Constructor, from a dynamic array
		 *\param[in]	p_pElements		The elements array
		 *\param[in]	p_uiNbElements	The count of elements
		 */
		BufferDeclaration( const BufferElementDeclaration * p_pElements, size_t p_uiNbElements);
		/**
		 *\brief	Destructor
		 */
		virtual ~BufferDeclaration();
		/**
		 *\return	An iterator over the beginning of the array
		 */
		inline iterator Begin() { return m_arrayElements.begin(); }
		/**
		 *\return	A constant iterator over the beginning of the array
		 */
		inline const_iterator Begin()const { return m_arrayElements.begin(); }
		/**
		 *\return	A constant iterator over the end of the array
		 */
		inline const_iterator End()const { return m_arrayElements.end(); }
		/**
		 *\return	The number of elements
		 */
		inline size_t Size()const { return m_arrayElements.size(); }
		/**
		 *\return	The stride
		 */
		inline size_t GetStride()const { return m_uiStride; }

	private:
		void DoInitialise( const BufferElementDeclaration * p_pElements, size_t p_uiNbElements);
	};

	class BufferElement
	{
	protected:
		typedef PointBase element;							//!< Typedef over a \p PointBase
		typedef element * element_ptr;						//!< Typedef over a pointer of \p element
		typedef std::vector<element> element_array;			//!< Typedef over an array of \p element
		typedef std::vector<element_ptr> element_ptr_array;	//!< Typedef over an array of \p element_ptr

		BufferDeclaration * m_pDeclaration;					//!< The declaration, used to define every part of the buffer element
		unsigned char * m_pBuffer;							//!< The buffer, containing buffer values
		element_ptr_array m_arrayElements;					//!< The array of elements, each of them pointing on a part of the buffer
		size_t m_uiStride;									//!< The stride of this element
		bool m_bOwnBuffer;									//!< Tells if this element owns it's buffer or not

	public:
		/**
		 *\brief		Constructor
		 *\remark		This BufferElement owns it's buffer
		 *\param[in]	p_pDeclaration	The declaration
		 */
		BufferElement( BufferDeclaration * p_pDeclaration );
		/**
		 *\brief		Destructor, deletes the buffer if owned by this element
		 */
		virtual ~BufferElement();
		/**
		 *\brief		Copy constructor
		 *\remark		This BufferElement doesn't own it's buffer
		 *\param[in]	p_element	The element to copy
		 */
		BufferElement( const BufferElement & p_element );
		/**
		 *\brief		Assignement operator
		 *\remark		This BufferElement doesn't own it's buffer anymore
		 *\param[in]	p_element		The element to copy
		 */
		BufferElement & operator =( const BufferElement & p_bufferElement);
		/**
		 *\brief		Links the element values to the ones in parameter. The element no longer owns it's values
		 *\param[in]	p_pBuffer	The values buffer
		 */
		void LinkCoords( void * p_pBuffer);
		/**
		 *\brief		Unlinks the element values. The element now owns it's coords
		 */
		void UnlinkCoords();
		/**
		 *\brief		Gets a reference to an element at given index
		 *\param[in]	p_uiIndex		The index
		 */
		template <typename T> inline T & GetElement( size_t p_uiIndex ) { return static_cast<T &>( * m_arrayElements[p_uiIndex]); }
		/**
		 *\brief		Gets a constant reference to an element at given index
		 *\param[in]	p_uiIndex		The index
		 */
		template <typename T> inline T const & GetElement( size_t p_uiIndex )const { return static_cast<T &>( * m_arrayElements[p_uiIndex]); }
		/**
		 *\return		Gets the element stride (size of an element)
		 */
		inline size_t GetStride()const { return m_uiStride; }
		/**
		 *\return		A pointer to the buffer values
		 */
		inline unsigned char * ptr() { return m_pBuffer; }
		/**
		 *\return		A constant pointer to the buffer values
		 */
		inline const unsigned char * const_ptr()const { return m_pBuffer; }

	protected:
		void DoLink();
	};
}

#endif
