/*
This source file is part of ProceduralGenerator (https://sourceforge.net/projects/proceduralgene/)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___PG_Plugin___
#define ___PG_Plugin___

#if defined( __WXMSW__ )
#	include <process.h>
#	define DECLARE_PLUGIN() virtual void Destroy();
#	define IMPLEMENT_PLUGIN( plugin, generator)	extern "C"\
												{\
													__declspec( dllexport) PluginBase * CreatePlugin( wxDynamicLibrary * p_pDll, wxApp * p_pApp ){ return Plugin< generator, plugin >::Creator( p_pDll, p_pApp ); }\
													__declspec( dllexport) void DestroyPlugin() { Plugin< generator, plugin >::Destroyer(); }\
												}\
												IMPLEMENT_APP_NO_MAIN( plugin)\
												void plugin::Destroy()\
												{\
													DestroyPlugin();\
												}
#else
#	define DECLARE_PLUGIN() virtual void Destroy();
#	define IMPLEMENT_PLUGIN( plugin, generator)	extern "C" PluginBase * CreatePlugin( wxDynamicLibrary * p_pDll, wxApp * p_pApp ) { return new plugin; }\
												void plugin::Destroy()\
												{\
													delete this;\
												}
#endif

namespace ProceduralTextures
{
	class PluginBase
#if defined( __WXMSW__ )
        :   public wxApp
#endif
	{
	public:
		PluginBase();
		virtual ~PluginBase();

		virtual void Destroy()=0;

		virtual ProceduralGenerator * CreateGenerator( int p_iWidth, int p_iHeight, int p_iWndId, wxFrame * p_pFrame)=0;
		virtual wxString GetName()=0;
	};

	template< class GeneratorClass, class PluginClass >
	class Plugin : public PluginBase
	{
#if defined( __WXMSW__ )
	private:
		static wxCriticalSection	m_critStartup;
		static HANDLE				m_hThreadId;
		static wxDynamicLibrary	*	m_pDll;

		typedef enum
		{	eTerminate
		}eIDs;

		static unsigned int wxSTDCALL Launcher( void * p_event )
		{
#	if wxVERSION_NUMBER < 2900
			const HINSTANCE l_hInstance = m_pDll->GetLibHandle();
#	else
			const HINSTANCE l_hInstance = wxDynamicLibrary::MSWGetModuleHandle( "", & Plugin< GeneratorClass, PluginClass >::m_hThreadId);
#	endif

			if ( ! l_hInstance )
			{
				return 0;
			}

#	if wxVERSION_NUMBER >= 2900
			wxDISABLE_DEBUG_SUPPORT();
#	endif
			wxInitializer l_wxinit;

			if ( ! l_wxinit.IsOk() )
			{
				return 0;
			}

			HANDLE l_hEvent = *(static_cast<HANDLE *>(p_event));

			if ( ! ::SetEvent( l_hEvent) )
			{
				return 0;
			}

			wxEntry( l_hInstance);
			return 1;
		}

	public:
		static PluginBase * Creator( wxDynamicLibrary * p_pDll, wxApp * p_pApp )
		{
			PluginBase * pReturn = NULL;
			wxCriticalSectionLocker lock( m_critStartup );

			if ( ! m_hThreadId )
			{
				m_pDll = p_pDll;
				HANDLE l_hEvent = ::CreateEvent( NULL, FALSE, FALSE, NULL);

				if ( ! l_hEvent )
				{
					return NULL;
				}

				m_hThreadId = (HANDLE)_beginthreadex( NULL, 0, & Launcher, & l_hEvent, 0, NULL );

				if ( ! m_hThreadId )
				{
					::CloseHandle( l_hEvent);
					return NULL;
				}

				::WaitForSingleObject( l_hEvent, INFINITE);
				::CloseHandle( l_hEvent);
				pReturn = (PluginBase *)wxApp::GetInstance();
			}

			return pReturn;
		}
		static void Destroyer()
		{
			wxCriticalSectionLocker lock( m_critStartup);

			if ( ! m_hThreadId )
			{
				return;
			}

			m_pDll->Detach();
			delete m_pDll;

			wxThreadEvent * l_pEvent = new wxThreadEvent( wxEVT_THREAD, eTerminate);
			wxTheApp->AddPendingEvent( * l_pEvent);

			// We must then wait for the thread to actually terminate.
			if( ::WaitForSingleObject( m_hThreadId, 10000 ) == WAIT_TIMEOUT )
			{
				::TerminateThread( m_hThreadId, 0xFFFFFFFF );
			}
			CloseHandle( m_hThreadId);
			m_hThreadId = NULL;
			delete l_pEvent;
		}
		void OnTerminate( wxThreadEvent & event)
		{
			ExitMainLoop();
		}
#endif
	public:
		Plugin()
			:	PluginBase()
		{
#if defined( __WXMSW__ )
			Connect( eTerminate, wxEVT_THREAD, wxThreadEventHandler( Plugin::OnTerminate));
#endif
		}
		virtual ~Plugin()
		{
		}
		ProceduralGenerator * CreateGenerator( int p_iWidth, int p_iHeight, int p_iWndId, wxFrame * p_pFrame )
		{
			return new GeneratorClass( p_iWidth, p_iHeight, p_iWndId, p_pFrame);;
		}
		bool OnInit()
		{
			return true;
		}
		int OnExit()
		{
			return 0;
		}
	};

	typedef PluginBase * (CreatePluginFunction)( wxDynamicLibrary * p_pDll, wxApp * p_pApp );
	typedef CreatePluginFunction * PCreatePluginFunction;
#if defined( __WXMSW__ )
	template< class GeneratorClass, class PluginClass > wxCriticalSection	Plugin< GeneratorClass, PluginClass >::m_critStartup;
	template< class GeneratorClass, class PluginClass > HANDLE				Plugin< GeneratorClass, PluginClass >::m_hThreadId		= NULL;
	template< class GeneratorClass, class PluginClass > wxDynamicLibrary *	Plugin< GeneratorClass, PluginClass >::m_pDll			= NULL;
#endif
}

#endif
