/*
This source file is part of ProceduralGenerator (https://sourceforge.net/projects/proceduralgene/)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___ProceduralGenerator___
#define ___ProceduralGenerator___

#include "Pixel.h"
#include "ConfigPanel.h"
#include "Plugin.h"
#include "GlObjects.h"

namespace ProceduralTextures
{
	class ProceduralGenerator : public wxThread, public wxEvtHandler
	{
	protected:
		/*!
		\author Sylvain DOREMUS
		\date 23/05/2012
		\~english
		\brief	Worker thread
		\~french
		\brief	Thread de travail
		*/
		class Thread : public wxThread
		{
		private:
			bool					m_bStopped;
			bool					m_bLaunched;

		protected:
			mutable wxMutex			m_mutex;
			ProceduralGenerator	*	m_pParent;
			int						m_iWidth;
			int						m_iTop;
			int						m_iBottom;
			int						m_iHeight;
			UbPixel					m_pxColour;
			size_t					m_uiIndex;

		public:
			Thread( ProceduralGenerator * p_pParent, size_t p_uiIndex, int iWidth, int iTop, int iBottom, int iTotalHeight, const UbPixel & p_pxColour );
			virtual ~Thread();

			virtual ExitCode Entry();

			virtual void Step()=0;

			inline bool IsStopped	()const					{ wxMutexLocker l_locker( m_mutex );return m_bStopped;		}
			inline void Stop		()						{ wxMutexLocker l_locker( m_mutex );m_bStopped = true;		}
			inline bool IsLaunched	()const					{ wxMutexLocker l_locker( m_mutex );return m_bLaunched;		}
			inline void SetLaunched	( bool p_bVal )			{ wxMutexLocker l_locker( m_mutex );m_bLaunched = p_bVal;	}
			inline void SetRed		( unsigned char val)	{ m_pxColour.r = val;										}
			inline void SetGreen	( unsigned char val)	{ m_pxColour.g = val;										}
			inline void SetBlue		( unsigned char val)	{ m_pxColour.b = val;										}
		};

	protected:
		wxSize						m_sizeDisplay;
		wxSize						m_sizeImage;
		unsigned long long			m_step;
		PixelBuffer					m_frontBuffer;
		PixelBuffer					m_backBuffer;
		UbPixel						m_pxColour;
		size_t						m_uiThreadCount;
		size_t						m_uiThreadEndedCount;
		bool						m_bStopped;
		wxFrame *					m_pMainFrame;
		bool						m_bNeedsConfigPanel;
		std::vector< Thread * >		m_arraySlaveThreads;
		bool						m_initialised;
		GlVertexBuffer				m_vertexBuffer;
		GlIndexBuffer				m_indexBuffer;
		GlFrameBuffer				m_frameBuffer;
		GlTexture					m_texture;
		GlFrameBuffer				m_frameBufferSave;
		GlTexture					m_textureSave;
		bool						m_bToSave;
		mutable wxMutex				m_mutex;
		bool						m_bRendered;
		wxObjectEventFunction		m_pfnCpuStepEndHandler;
		CtrlInfosArray				m_arrayControls;
		int							m_iSleepTime;

	public:
		ProceduralGenerator( int p_width, int p_height, int p_iWndId, wxFrame * p_pMainFrame, bool p_bNeedsConfigPanel=true, int p_iSleepTime=10 );
		virtual ~ProceduralGenerator();

		virtual ExitCode		Entry				();
		virtual void			SetRed				( unsigned char val);
		virtual void			SetGreen			( unsigned char val);
		virtual void			SetBlue				( unsigned char val);
		virtual void			SwapBuffers			();
		virtual void			InitialiseStep		();
		virtual void			ClearAllThreads		();
		virtual void			GlInit				();
		virtual void			GlPreRender			();
		virtual void			GlRender			( bool p_bChanged);
		virtual void			GlPostRender		();
		virtual void			GlCleanup			();

		inline void SetCpuStepEndHandler( wxObjectEventFunction p_pfnHandler ) { m_pfnCpuStepEndHandler = p_pfnHandler; }

		void WaitForStepEnd();
		bool AllEnded();
		void Suspend();
		void SetSize( int p_iWidth, int p_iHeight);
		void SetSize( const wxSize & p_size);
		void Resize( const wxSize & p_size);
		void SaveFrame();

		template <class ThreadClass>
		ThreadClass * CreateThread( int iWidth, int iTop, int iBottom, int iTotalHeight, const UbPixel & p_pxColour)
		{
			ThreadClass * l_pThread = new ThreadClass( this, m_arraySlaveThreads.size(), m_sizeImage.x, iTop, iBottom, iTotalHeight, p_pxColour);
			m_arraySlaveThreads.push_back( l_pThread);
			l_pThread->Create();
			return l_pThread;
		}

		inline unsigned char					GetRed				()const				{ return m_pxColour.r;														}
		inline unsigned char					GetGreen			()const				{ return m_pxColour.g;														}
		inline unsigned char					GetBlue				()const				{ return m_pxColour.b;														}
		inline bool								IsInitialised		()const				{ return m_initialised;														}
		inline const UbPixel *					GetPixels			()const				{ return m_frontBuffer.const_ptr();											}
		inline const PixelBuffer *				GetPixelsBuffer		()const				{ return & m_frontBuffer;													}
		inline PixelBuffer *					GetPixelsBuffer		()					{ return & m_frontBuffer;													}
		inline wxSize const &					GetDisplaySize		()const				{ return m_sizeDisplay;														}
		inline wxSize const &					GetImageSize		()const				{ return m_sizeImage;														}
		inline bool								NeedsConfigPanel	()const				{ return m_bNeedsConfigPanel;												}
		inline wxFrame *						GetMainFrame		()const				{ return m_pMainFrame;														}
		inline bool								IsRendered			()const				{ wxMutexLocker l_locker( m_mutex );return m_bRendered;						}
		inline void								SetRendered			( bool p_bVal )		{ wxMutexLocker l_locker( m_mutex );m_bRendered = p_bVal;					}
		inline bool								IsStopped			()const				{ wxMutexLocker l_locker( m_mutex );return m_bStopped;						}
		inline void								Stop				()					{ wxMutexLocker l_locker( m_mutex );m_bStopped = true;ClearAllThreads();	}
		inline CtrlInfosArray::iterator			BeginCtrls			()					{ return m_arrayControls.begin();											}
		inline CtrlInfosArray::iterator			EndCtrls			()					{ return m_arrayControls.end();												}

	protected:
		virtual void DoGeneratePanel	()=0;
		virtual void DoResize			( const wxSize & p_size);
		virtual void DoGlInit			()=0;
		virtual void DoGlPreRender		()=0;
		virtual void DoGlRender			( bool & p_bChanged)=0;
		virtual void DoGlPostRender		()=0;
		virtual void DoGlCleanup		()=0;

		void	DoCleanup	();
		wxSize	DoLoadImage	( const wxString & p_strImagePath, wxImage & p_pImage, PixelBuffer & p_buffer);
		void	DoSubRender	();
		void	DoSaveFrame	();

	private:
		void OnThreadEnd	( wxCommandEvent & p_event );
		void OnStop			( wxCommandEvent & p_event );
	};
}

#endif
