/*
This source file is part of ProceduralGenerator (https://sourceforge.net/projects/proceduralgene/)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___Pg_Vertex___
#define ___Pg_Vertex___

#include "BufferElement.h"
#include "Point.h"

namespace ProceduralTextures
{
	class Vertex : public BufferElement
	{
	public:
		static size_t Size;					//!< Number of components of a Vertex (total of coordinates for coords, normal, tangent, texture coords).
		static unsigned long long Count;	//!< Total number of created vertex

	private:
		size_t m_uiIndex;
		bool m_bManual;

	public:
		/**
		* Specified constructor
		*/
		Vertex( BufferDeclaration * p_pVtxDecl, float x = 0, float y = 0, float z = 0, size_t p_uiIndex = 0);
		/**
		* Copy constructor
		*/
		Vertex( const Vertex & p_vertex );
		/**
		* Constructor from point
		*/
		Vertex( BufferDeclaration * p_pVtxDecl, const Point3f & p_point);
		/**
		* Constructor from a real array
		*/
		Vertex( BufferDeclaration * p_pVtxDecl, float * p_pBuffer);
		/**
		* Constructor from the difference between second and first argument
		*/
		Vertex( BufferDeclaration * p_pVtxDecl, const Vertex & p_v1, const Vertex & p_v2);
		/**
		* Destructor
		*/
		virtual ~Vertex();
		/**
		 * Links the vertex coords to the ones in parameter.
		 * The vertex no longer owns it's coords
		 *@param p_pBuffer : [in] The coordinates buffer
		 */
		void LinkCoords( float * p_pBuffer);
		/**
		 * Unlinks the vertex coords.
		 * The vertex now owns it's coords
		 */
		void UnlinkCoords();

		/**@name Assignment */
		//@{
		void SetNormal	( const Point3f & val, bool p_bManual = false);
		void SetNormal	( float x, float y, float z, bool p_bManual = false);
		void SetNormal	( const float * p_pCoords, bool p_bManual = false);

		void SetTexCoord	( const Point2f & val, bool p_bManual = false);
		void SetTexCoord	( float x, float y, bool p_bManual = false);
		void SetTexCoord	( const float * p_pCoords, bool p_bManual = false);

		void SetTangent	( const Point3f & val, bool p_bManual = false);
		void SetTangent	( float x, float y, float z, bool p_bManual = false);
		void SetTangent	( const float * p_pCoords, bool p_bManual = false);
		//@}

		/**@name Operators */
		//@{
		inline float &			operator []( size_t p_uiIndex)		{ return GetElement<Point3f>( 0)[p_uiIndex]; }
		inline const float &	operator []( size_t p_uiIndex)const	{ return GetElement<Point3f>( 0)[p_uiIndex]; }
		virtual Vertex & 		operator =( const Vertex & p_vertex);
		//@}

		/**@name Accessors */
		//@{
		inline Point3f &		GetCoords	()		{ return GetElement<Point3f>( 0); }
		inline Point3f &		GetNormal	()		{ return GetElement<Point3f>( 1); }
		inline Point3f &		GetTangent	()		{ return GetElement<Point3f>( 2); }
		inline Point2f &		GetTexCoord	()		{ return GetElement<Point2f>( 3); }
		inline const Point3f &	GetCoords	()const	{ return GetElement<Point3f>( 0); }
		inline const Point3f &	GetNormal	()const	{ return GetElement<Point3f>( 1); }
		inline const Point3f &	GetTangent	()const	{ return GetElement<Point3f>( 2); }
		inline const Point2f &	GetTexCoord	()const	{ return GetElement<Point2f>( 3); }
		inline const float *	const_ptr	()const	{ return (const float *)( BufferElement::const_ptr()); }
		inline float *			ptr			()		{ return (float *)( BufferElement::ptr()); }
		inline size_t			GetIndex	()const { return m_uiIndex; }
		inline bool				IsManual	()const	{ return m_bManual; }

		inline void SetCoords	( const Point3f & val)			{ GetElement<Point3f>( 0) = val; }
		inline void SetCoords	( float x, float y, float z)	{ GetElement<Point3f>( 0)[0] = x;GetElement<Point3f>( 0)[1] = y;GetElement<Point3f>( 0)[2] = z; }
		inline void SetCoords	( const float * p_pCoords)		{ GetElement<Point3f>( 0)[0] = GetElement<Point3f>( 0)[0];GetElement<Point3f>( 0)[1] = p_pCoords[1];GetElement<Point3f>( 0)[2] = p_pCoords[2]; }
		inline void SetIndex	( const size_t & val)			{ m_uiIndex = val; }
		//@}

	private:
		void DoLink();
	};
}

#endif
